# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Operations on graphs."""
from functools import reduce
from typing import Set

from rdflib import Graph
from rdflib.term import URIRef

from gimie.graph import Property


def combine_graphs(*graphs: Graph) -> Graph:
    """Combines an arbitrary number of input graphs
    into a single graph."""
    return reduce(lambda g1, g2: g1 | g2, graphs)


def properties_to_graph(uri: URIRef, properties: Set[Property]) -> Graph:
    """Attaches a set of predicate-object tuples to input
    URI to produce an RDF graph."""
    g = Graph()
    for pred, obj in properties:
        g.add((uri, pred, obj))
    return g
